/**********************************************************************************/
/*                                                                                */
/* File:   bsp.h                                                                  */
/* Author: bkenwright@xbdev.net                                                   */
/* URL:    www.xbdev.net                                                          */
/* Date:   22-12-2005 (xmas)                                                      */
/*                                                                                */
/**********************************************************************************/

#ifndef BSP_H
#define BSP_H

#include <d3dx9.h>								// DirectX Header Files
#pragma comment(lib, "d3d9.lib")				// DirectX Library Files
#pragma comment(lib, "d3dx9.lib")				// IDirect3DDevice* etc

#include "common.h"								// SafeRelease(..)

#include "dxdbg.h"								// Triangle3D(..)


//----------------------- Various BSP Lump Data Types-----------------------------//
enum BSP_TYPES
{
	enEntities=0,
	enTextures,
	enPlanes,
	enNodes,
	enLeaves,
	enLeafFaces,
	enLeafBrushes,
	enModels,
	enBrushes,
	enBrushSides,
	enVertices,
	enMeshIndices,
	enEffect,
	enFaces,
	enLightmaps,
	enLightVols,
	enVisData
};

extern char szBSPTYPES [17][20];				// Defined in bsp.cpp

//char szBSPTYPES [17][20] = { 	
//	"Entities",
//	"Textures",
//	"Planes",
//	"Nodes",
//	"Leaves",
//	"LeafFaces",
//	"LeafBrushes",
//	"Models",
//	"Brushes",
//	"BrushSides",
//	"Vertices",
//	"MeshIndices",
//	"Effect",
//	"Faces",
//	"Lightmaps",
//	"LightVols",
//	"VisData" };


//------------------------- Various Structure Definitions ------------------------//
struct stLump
{
	int		nFileofs;			// Offset to start of lump, relative to beginning of file.
	int		nFileLen;			// Length of lump. Always a multiple of 4.
};

struct stHeader
{
	char   cMagic[4]; 
	int	   nVersion;
	stLump Lumps[17];
};

//----------------------------- Lump Definitions ---------------------------------//

typedef float	Vector2[2];

struct Vector3
{
	union
	{
		struct {float x,y,z;};
		float m[3];
	};
	float operator [](int i){ return m[i]; }

	float Dot(const Vector3 & v)
	{
		return (x*v.x + y*v.y + z*v.z);
	}

	float Dot(const D3DXVECTOR3 & v)
	{
		return (x*v.x + y*v.y + z*v.z);
	}
};

typedef float	Vector4[4];

typedef float	TexCoord[2];

typedef int		nBBox[6]; // Integer bounding box (mins, maxs)

typedef float	fBBox[6]; // Float bounding box
 

//Lump 0
//Entities
//char *entities; // A pointer to text
 

//Lump 1
//Shader Texture Info
struct stTexture
{
	char	Name[64];		// Texture name
	int		nSurfaceFlags;	// Type of surface (See Surface Flags below)
	int		nContentFlags;	// Leaf content (See Content Flags below)
};
 

//Lump 2
//Planes
struct stPlane 
{
	Vector3 Normal;			// Normal vector for plane
	float	fDist;			// Distance from plane to origin
};
 

//Lump 3
//Nodes
struct stNode 
{
	int		nPlane;			// Space partitioning plane
	int		nChildren[2];	// Back and front child nodes
	int		BBoxIMin[3];	// Bounding box of node
	int		BBoxIMax[3];	// Bounding box of node
};
 
//Lump 4
//Leaves 
struct stLeaf
{
	int		nCluster;		// Visibility cluster number
	int		nArea;			// Volume of the leaf
	int		BBoxIMin[3];	// Bounding box of leaf
	int		BBoxIMax[3];	// Bounding box of leaf

	int		nFirstFace, 
			NumFaces;		// Lookup for the face list (indexes 
							// are for faces)

	int		nFirstBrush, 
			NumBrushes;		// Lookup for the brush list (indexes
							// are for brushes)
};
 

//Lump 5
//Leaf Faces
//int *pFaces;				// a pointer to a series of indexes to 
							// a face list
 
//Lump 6
//Leaf Brushes
//int *pBrushes;			// a pointer to a series of indexes to 
							// a brush list
 
//Lump 7
//Models
struct stModel 
{
	fBBox	BBoxF;			// Bounding box of model

	int		nFirstFace,		// First face for model
			NumFaces;		// Number of faces for model 

	int		nFirstBrush,	// First brush for model
			NumBrushes;		// Number of brushes for model
};
 
//Lump 8
//Brushes
struct stBrush 
{
	int		nFirstSide,		// First brushside for brush
			NumSides;		// Number of brushsides for brush

	int		nIndex;			// Texture index
};
 
//Lump 9
//Brush Sides
struct stBrushSide 
{
	int		PlaneNum;		// Lookup for plane
	int		nIndex;			// Texture index
};
 
//Lump 10
//Vertices
struct stVertex 
{
	Vector3  vPoint;		// Vertex Position
	TexCoord Tex;			// Texture coordinates
	TexCoord LightTexCoord; // Light Map texture coordinates
	Vector3  vNormal;		// Normal vector (used for lighting ?)

	unsigned int RGBA;		// Vertex color. RGBA
};
 

//Lump 11
//MeshVert 
//int nOffset;				// Vertex index offset, relative to first 
							// vertex of corresponding face.  

 

//Lump 12
//Effect 
struct stEffect
{
	char cName[64];			// Effect shader.  
	int	 nBrush;			// Brush that generated this effect.  
	int  Unknown;			// Always 5, except in q3dm8, which has 
							// one effect with -1.  
};


 
//Lump 13
//Faces
#pragma pack(1)
struct stFace // size 24*4
{
	int nTexture;			// Refers to a texture
	int nEffect;			// Index into lump 12 (Effects), or -1
	int nFaceType;			// Face type. 1=polygon, 2=patch, 3=mesh, 4=billboard 
	int FirstVert, 
		NumVerts;			// Reference to vertices

	int nFirstMeshVerts,	// Index of first meshvert
							// Every three meshverts describe a triangle
		NumMeshVerts;		// Number of meshverts

	int LMIndex;			// Lightmap index.
	int LMStart[2];			// X,Y Corner of this face's lightmap image in lightmap.
	int LMSize[2];			// Size of lightmap

	float LMOrigin[3];		// World space origin of lightmap.
	float LMVects[2][3];	// World space lightmap s and t unit vectors.

	Vector3 vNormal;		// Face normal

	int nSize[2];			// Patch dimensions.  
};
#pragma pack()


//Lump 14
//Lightmaps
//unsigned char pMap[128][128][3]; // Lightmap color data. RGB.  
 
//Lump 15
//Light Grid
//Unknown
 
//Lump 16
//Visibility Lists
struct stVisData 
{
	int NumVectors;			// Number of vectors.
	int nSizeVectorInBytes;	// Size of each vector, in bytes
	unsigned char *pBits;	// [NumVectors * nSizeVector];  
							// Visibility data. One bit per cluster per vector
};

//Lump 17
//Number of Lumps
//Unknown
 

void ErrLog(char *str);		// bsp.cpp


class CBSP
{
	stHeader	m_Header;

	stVertex    *m_pVerts;
	int			m_nNumVerts;

	stFace		*m_pFaces;
	int			m_nNumFaces;

	int			*m_pMeshIndices;
	int			m_nNumMeshIndices;

	stTexture   *m_pTextures;
	int			m_nNumTextures;

	stNode		*m_pNodes;
	int			m_nNumNodes;

	stLeaf		*m_pLeafs;
	int			m_nNumLeafs;

	stPlane		*m_pPlanes;
	int			m_nNumPlanes;

	int			*m_pLeafFaces;
	int			m_nNumLeafFaces;

	stVisData	m_VisData;
	int			m_nNumVisData;

	// Used for DirectX Vertex Buffer
	struct stBSPVertex
	{
		D3DXVECTOR3 p;			// Position Vertice
		D3DXVECTOR3 n;			// Vertice Normal
		DWORD		colour;		// ARGB
		float       tu1,tv1;	// Texture Coordinates
		float		lu2,lv2;	// Light Tex Coords
	};

	IDirect3DVertexBuffer9 * m_pVB;
	UINT		m_FVF_Format;

	IDirect3DIndexBuffer9* m_pIB;

	IDirect3DTexture9 ** m_pT;

	// Keep track of the actual number draw, not the actual number
	// in the total.
	int			m_nNumTrisDraw;

	// An array the same size as m_nNumFaces, so we can
	// keep track of which faces we need to render, instead of
	// rendering them twice.
	bool		*m_bFaceVisibility;

public:

	CBSP(IDirect3DDevice* pDevice, char * szFile )
	{ 
		// Always initialize variables
		m_nNumVerts			= 0;
		m_nNumFaces			= 0;
		m_nNumMeshIndices	= 0;
		m_nNumTextures		= 0;
		m_nNumNodes			= 0;
		m_nNumLeafs			= 0;
		m_nNumPlanes		= 0;
		m_nNumLeafFaces		= 0;
		m_nNumVisData		= 0;
		m_nNumTrisDraw		= 0;

		// Always initialise pointers to NULL
		m_pFaces			= NULL;
		m_pVerts			= NULL;
		m_pMeshIndices		= NULL;
		m_pTextures			= NULL;
		m_pNodes			= NULL;
		m_pLeafs			= NULL;
		m_pPlanes			= NULL;
		m_pLeafFaces		= NULL;
		m_VisData.pBits		= NULL;
		m_bFaceVisibility   = NULL;

		m_pVB				= NULL;
		m_pIB				= NULL;
		m_pT				= NULL;

		Create(pDevice, szFile); 
	}
	~CBSP(){ Release(); }


	bool Release()
	{
		delete[] m_pVerts;
		m_pVerts=NULL;

		delete[] m_pFaces;
		m_pFaces=NULL;

		delete[] m_pMeshIndices;
		m_pMeshIndices=NULL;

		delete[] m_pTextures;
		m_pTextures=NULL;
		
		delete[] m_pNodes;
		m_pNodes=NULL;

		delete[] m_pLeafs;
		m_pLeafs=NULL;

		delete[] m_pPlanes;
		m_pPlanes=NULL;

		delete[] m_pLeafFaces;
		m_pLeafFaces=NULL;

		delete[] m_VisData.pBits;
		m_VisData.pBits		 =NULL;
		m_VisData.NumVectors =0;
		m_VisData.nSizeVectorInBytes=0;

		delete[] m_bFaceVisibility;
		m_bFaceVisibility=NULL;

		SafeRelease(m_pVB);
		SafeRelease(m_pIB);

		if( m_pT!=NULL )
		{
			for(int i=0; i<m_nNumTextures; i++)
			{
				SafeRelease(m_pT[i]);
			}
			delete[] m_pT;
			m_pT=NULL;
		}

		m_nNumVerts = 0;
		m_nNumFaces = 0;
		m_nNumMeshIndices = 0;
		m_nNumTextures = 0;
		m_nNumNodes	= 0;
		m_nNumLeafs = 0;
		m_nNumPlanes = 0;
		m_nNumLeafFaces = 0;
		m_nNumVisData = 0;
		m_nNumTrisDraw = 0;
		return true;					// No Error
	}

//------------------------------Some statistic data-------------------------------//
	int GetTotalTris()
	{
		// 3 indices make up a triangle
		return m_nNumMeshIndices/3;
	}

	int GetDrawnTris()
	{
		// Since we only call GetDrawnTris() usually once per loop, seems okay
		// to clear its value after we have called it
		int nNumTrisDraw = m_nNumTrisDraw;
		ClearTriCounter();
		return nNumTrisDraw;
	}

	void ClearTriCounter()
	{
		m_nNumTrisDraw=0;
	}

//------------------------------------Swizzle-------------------------------------//

	//Vector3 SwizzleYZ(Vector3 & vIN)
	//{
	//	Vector3 vOUT = vIN;
	//	vOUT.x =  vIN.x;
	//	vOUT.y =  vIN.z;
	//	vOUT.z = -vIN.y;
	//	return vOUT;
	//}

//------------------------------------Vertices------------------------------------//
	bool LoadVerticeData(FILE* fp,
						 stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading Vertices Data\n");

		stLump * pLump = &pHeaderLump[enVertices];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pVerts = (stVertex*)new char[pLump->nFileLen];
		assert(m_pVerts);

		fread(	(void*)m_pVerts,		// Buffer
				1,						// Size
				pLump->nFileLen,		// Number of times
				fp );					// File Pointer

		
		m_nNumVerts = pLump->nFileLen / sizeof(stVertex); 

		// Record Information
		char buf[100];					// Temp text buffer
		sprintf(buf, "<?>Num Vertices: %d\n", m_nNumVerts);
		ErrLog(buf);

		ErrLog("<->End Loading Vertices Data\n");
		return true;
	}

//------------------------------------ Faces -------------------------------------//
	bool LoadFacesData(FILE* fp,
					   stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading Faces Data\n");

		stLump * pLump = &pHeaderLump[enFaces];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pFaces = (stFace*)new char[pLump->nFileLen];
		assert(m_pFaces);

		fread(	(void*)m_pFaces,		// Buffer
				pLump->nFileLen,		// Size
				1,						// Number of times
				fp );					// File Pointer


		m_nNumFaces = pLump->nFileLen / sizeof(stFace); 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num Faces: %d\n", m_nNumFaces);
		ErrLog(buf);

		ErrLog("<->End Loading Vertices Data\n");
		return true;
	}

//-------------------------------- Mesh Vert Tris -------------------------------------//
	bool LoadMeshVertsData(	FILE* fp,
							stLump * pHeaderLump)
	{
		ErrLog("<->Start MeshVert Data\n");

		stLump * pLump = &pHeaderLump[enMeshIndices];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pMeshIndices = (int*)new char[pLump->nFileLen];
		assert(m_pMeshIndices);

		fread(	(void*)m_pMeshIndices,	// Buffer
				pLump->nFileLen,		// Size
				1,						// Number of times
				fp );					// File Pointer


		m_nNumMeshIndices = pLump->nFileLen / sizeof(int); 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num Mesh Verts: %d\n", m_nNumMeshIndices);
		ErrLog(buf);

		ErrLog("<->End MeshVert Data\n");
		return true;
	}

	bool LoadTextureData(	FILE* fp,
							stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading Texture Data\n");

		stLump * pLump = &pHeaderLump[enTextures];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pTextures = (stTexture*)new char[pLump->nFileLen];
		assert(m_pTextures);

		fread(	(void*)m_pTextures,		// Buffer
				pLump->nFileLen,		// Size
				1,						// Number of times
				fp );					// File Pointer


		m_nNumTextures = pLump->nFileLen / sizeof(stTexture); 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num Textures: %d\n", m_nNumTextures);
		ErrLog(buf);

		ErrLog("<->End Loading Texture Data\n");

		return true;
	}


	bool LoadNodesData(	FILE* fp,
						stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading Nodes Data\n");

		stLump * pLump = &pHeaderLump[enNodes];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pNodes = (stNode*)new char[pLump->nFileLen];
		assert(m_pNodes);

		fread(	(void*)m_pNodes,		// Buffer
				pLump->nFileLen,		// Size
				1,						// Number of times
				fp );					// File Pointer


		m_nNumNodes = pLump->nFileLen / sizeof(stNode); 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num Nodes: %d\n", m_nNumNodes);
		ErrLog(buf);

		ErrLog("<->End Loading Nodes Data\n");
		return true;
	}

	bool LoadLeafsData(	FILE* fp,
						stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading Leafs Data\n");

		stLump * pLump = &pHeaderLump[enLeaves];

		m_nNumLeafs = pLump->nFileLen / sizeof(stLeaf);

		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pLeafs = (stLeaf*)new char[pLump->nFileLen];
		assert(m_pLeafs);

		fread(	(void*)m_pLeafs,		// Buffer
				sizeof(stLeaf),			// Size
				m_nNumLeafs,			// Number of times
				fp );					// File Pointer


		 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num Leafs: %d\n", m_nNumLeafs);
		ErrLog(buf);

		ErrLog("<->End Loading Leafs Data\n");
		return true;
	}

	bool LoadPlanesData( FILE* fp,
						 stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading Planes Data\n");

		stLump * pLump = &pHeaderLump[enPlanes];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pPlanes = (stPlane*)new char[pLump->nFileLen];
		assert(m_pPlanes);

		fread(	(void*)m_pPlanes,		// Buffer
				pLump->nFileLen,		// Size
				1,						// Number of times
				fp );					// File Pointer


		m_nNumPlanes = pLump->nFileLen / sizeof(stPlane); 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num Planes: %d\n", m_nNumPlanes);
		ErrLog(buf);

		ErrLog("<->End Loading Planes Data\n");
		return true;
	}

	bool LoadLeafFacesData( FILE* fp,
						 stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading LeafFaces Data\n");

		stLump * pLump = &pHeaderLump[enLeafFaces];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		m_pLeafFaces = (int*)new char[pLump->nFileLen];
		assert(m_pLeafFaces);

		fread(	(void*)m_pLeafFaces,	// Buffer
				pLump->nFileLen,		// Size
				1,						// Number of times
				fp );					// File Pointer


		m_nNumLeafFaces = pLump->nFileLen / sizeof(int); 

		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num LeafFaces: %d\n", m_nNumLeafFaces);
		ErrLog(buf);

		ErrLog("<->End Loading LeafFaces Data\n");
		return true;
	}

	bool LoadVisData( FILE* fp,
					  stLump * pHeaderLump)
	{
		ErrLog("<->Start Loading VisData Data\n");

		stLump * pLump = &pHeaderLump[enVisData];
		fseek(	fp,
				pLump->nFileofs,
				SEEK_SET );

		fread(	(void*)&m_VisData,		// Buffer
				sizeof(int),			// Size
				2,						// Number of times
				fp );					// File Pointer

		m_nNumVisData = m_VisData.NumVectors;

		int size = (m_VisData.NumVectors * m_VisData.nSizeVectorInBytes);

		m_VisData.pBits = new unsigned char[size];
		assert(m_VisData.pBits);

		

		fread((void*)(m_VisData.pBits),		// Buffer
		sizeof(byte),						// Size
		size,								// Number of times
		fp );								// File Pointer


		// Record Information
		char buf[200]; //temp txt buffer
		sprintf(buf, "<?>Num VisData: %d\n", m_nNumVisData);
		ErrLog(buf);

		ErrLog("<->End Loading VisData Data\n");
		return true;
	}


	bool Create(IDirect3DDevice * pDevice,
				const char * szBSPFile)
	{
		if( szBSPFile==NULL )
		{
			ErrLog("<->Error Invalid File Name in BSP::Create(..)");
			return false;
		}

		Release();

		ErrLog("\n<->Opening File\n");

		// Open BSP File
		FILE * fp = fopen(szBSPFile, "rb");

		char buf[200]; //temp txt buffer

		if( fp==NULL )
		{
			sprintf(buf, 
				    "\n<?>Unable to open file: %s\n", 
					szBSPFile);
			ErrLog(buf);
			return false;
		}

		// Temporary Buffer
		fread(&m_Header,				// Buffer
			1,							// Size
			sizeof(stHeader),			// Number of times
			fp);						// File Pointer

		// Get the pointer to our array of lumps
		stLump * pLumps = m_Header.Lumps;


		//--- Loading & Creating Visual Side of things
		if( !LoadVerticeData(fp, pLumps)	) return false;
		
		if( !LoadFacesData(fp, pLumps)		) return false;

		if( !LoadMeshVertsData(fp, pLumps)	) return false;

		if( !LoadTextureData(fp, pLumps)	) return false;

		if( !CreateDXVertexBuffer(pDevice)	) return false;

		if( !CreateDXIndexBuffer(pDevice)	) return false;

		if( !CreateDXTextures(pDevice)		) return false;
		
		//---- Loading BSP Partitioning Info

		if( !LoadNodesData(fp, pLumps)		) return false;
		
		if( !LoadLeafsData(fp, pLumps)		) return false;

		if( !LoadPlanesData(fp, pLumps)		) return false;

		if( !LoadLeafFacesData(fp, pLumps)	) return false;

		if( !LoadVisData(fp, pLumps)		) return false;

		//-----Keep track of which faces need to be drawn
		m_bFaceVisibility = new bool[m_nNumFaces];
		assert(m_bFaceVisibility);

		ErrLog("\n<->Closing File\n");
		// Close BSP File
		fclose(fp);

		return true;					// No Error
	}

	bool CreateDXTextures( IDirect3DDevice * pDevice )
	{
		HRESULT hr = S_OK;				// Error checking variable
		if( m_nNumTextures==0 )
		{
			ErrMsg("<->No Textures To Load\n");
			return true;
		}

		m_pT = new IDirect3DTexture9*[m_nNumTextures];

		for(int i=0; i<m_nNumTextures; i++)
		{
			if( m_pTextures[i].nContentFlags != 1 )
			{
				m_pT[i] = NULL;
				continue;
			}

			// Doesn't include the '.jpg' on the end, hence we must fix this
			char * str = NULL;
			int len = 0;
			// Make sure we have a valid string / error checking
			if( m_pTextures[i].Name==NULL )
			{
				m_pT[i]=NULL;
				continue;
			}

			len = (int)strlen(m_pTextures[i].Name);
			if( len<=1 )
			{
				m_pT[i]=NULL;
				continue;
			}
			
			// copy and add the .jpg to the end
			str = new char[len+5]; // +4 is for ".jpg"
			strcpy(str, m_pTextures[i].Name);
			strcat(str, ".jpg");

			IDirect3DTexture9* pTexture = NULL;
			hr =
			D3DXCreateTextureFromFile(	pDevice,				// DirectX Device
										str,					// File Name
										&pTexture );			// Loaded Texture
			if( hr==S_OK )
				m_pT[i] = pTexture;
			else
				m_pT[i] = NULL;

			delete[] str;
		}

		return true;
	}

	bool CreateDXIndexBuffer(IDirect3DDevice * pDevice)
	{
		HRESULT hr = S_OK;

		hr =
		pDevice->CreateIndexBuffer( m_nNumMeshIndices * sizeof(WORD),
									D3DUSAGE_WRITEONLY,
									D3DFMT_INDEX16,
									D3DPOOL_DEFAULT,
									&m_pIB,
									NULL);
		if( hr != S_OK )
		{
			ErrLog("<!>Error Creating Index Buffer/n");
			return false;
		}

		// We need to copy the data one at a time, as the data is loaded
		// in as int, but directx uses a word for each indices
		WORD * pIB;
		m_pIB->Lock(0, 
					m_nNumMeshIndices*sizeof(WORD), 
					(void**)&pIB, 
					0);
		for(int i=0; i<m_nNumMeshIndices; i++)
			pIB[i] = m_pMeshIndices[i];
		m_pIB->Unlock();

		return true;
	}

	bool CreateDXVertexBuffer(IDirect3DDevice * pDevice)
	{
		HRESULT hr = S_OK;

		m_FVF_Format = D3DFVF_XYZ | 
					   D3DFVF_NORMAL |
					   D3DFVF_DIFFUSE |
					   D3DFVF_TEX2 |
					   D3DFVF_TEXCOORDSIZE2(0) |
					   D3DFVF_TEXCOORDSIZE2(1);
		hr =
		pDevice->CreateVertexBuffer( m_nNumVerts * sizeof(stBSPVertex),
									 0,
									 m_FVF_Format,
									 D3DPOOL_MANAGED,
									 &m_pVB,
									 NULL );
		if( hr != S_OK )
		{
			ErrLog("<!>Error Creating Vertex Buffer/n");
			return false;
		}

		// Copy data into our dx buffer
		stBSPVertex *pVB;
		m_pVB->Lock(0,0, (void**)&pVB, 0);

		for(int i=0; i<m_nNumVerts; i++)
		{
			// The put the y ans z the wrong way around!
			pVB[i].p.x = m_pVerts[i].vPoint[0];
			pVB[i].p.y = m_pVerts[i].vPoint[1];
			pVB[i].p.z = m_pVerts[i].vPoint[2];

			// tv is inverted!
			pVB[i].tu1 = m_pVerts[i].Tex[0];
			pVB[i].tv1 = m_pVerts[i].Tex[1];
			pVB[i].lu2 = m_pVerts[i].LightTexCoord[0];
			pVB[i].lv2 = m_pVerts[i].LightTexCoord[1];

			pVB[i].n.x   = m_pVerts[i].vNormal[0];
			pVB[i].n.y   = m_pVerts[i].vNormal[1];
			pVB[i].n.z   = m_pVerts[i].vNormal[2];

			// Alpha hack - so our alpha value isn't 100% anymore
			pVB[i].colour =   0x10ffffff & m_pVerts->RGBA; // ARGB
		}
		m_pVB->Unlock();


		return true;
	}

	// Renders ALL the level map
	// Currently doesn't check if it renders the same face contents twice, as
	// different faces can point to the same data.
	void Render(IDirect3DDevice9 * pDevice)
	{
#if(1)
		// Used for debugging
		pDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
		pDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE); 
		pDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
		pDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
#endif
		pDevice->SetTextureStageState(0,D3DTSS_COLOROP,	D3DTOP_SELECTARG1);
		pDevice->SetTextureStageState(0,D3DTSS_COLORARG1,	D3DTA_TEXTURE);
		pDevice->SetTextureStageState(0,D3DTSS_COLORARG2,	D3DTA_DIFFUSE);
		
		pDevice->SetIndices(m_pIB);

		pDevice->SetStreamSource(0, m_pVB, 0, sizeof(stBSPVertex));

		pDevice->SetFVF( m_FVF_Format );


		for(int i=0; i<m_nNumFaces; i++)
		{
			// If its not a tri mesh
			if( m_pFaces[i].nFaceType != 1 ) continue;

			int nFirstF = m_pFaces[i].nFirstMeshVerts;
			int nNumF   = m_pFaces[i].NumMeshVerts;

			int nFirstV = m_pFaces[i].FirstVert;
			int nNumV   = m_pFaces[i].NumVerts;

			int nTexIndex = m_pFaces[i].nTexture;
#if(0)
			pDevice->SetTexture( 0, 
								 m_pT[nTexIndex] );
#endif

			int nNumTris = nNumF / 3;

			m_nNumTrisDraw += nNumTris;

			pDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST,		// Type
					                        nFirstV,				// BaseVertexIndex
											0,						// MinIndex
											nNumV,					// NumVertices
											nFirstF,				// StartIndex
											nNumTris );				// PrimitiveCount
		}
	}


	void RenderFaces(IDirect3DDevice9 * pDevice,
					 int nStartFace,
					 int nNumFaces)
	{

		pDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
		pDevice->SetTextureStageState(0,D3DTSS_COLOROP,	D3DTOP_SELECTARG2);
		pDevice->SetTextureStageState(0,D3DTSS_COLORARG1,	D3DTA_DIFFUSE);
		pDevice->SetTextureStageState(0,D3DTSS_COLORARG2,	D3DTA_TEXTURE);
		
		pDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
		pDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE); 

		pDevice->SetIndices(m_pIB);

		pDevice->SetStreamSource(0, m_pVB, 0, sizeof(stBSPVertex));

		pDevice->SetFVF( m_FVF_Format );


		for(int i=nStartFace; i<nStartFace+nNumFaces; i++)
		{
			// If its not a tri mesh
			if( m_pFaces[i].nFaceType != 1)continue;

			int nFirstF = m_pFaces[i].nFirstMeshVerts;
			int nNumF   = m_pFaces[i].NumMeshVerts;

			int nFirstV = m_pFaces[i].FirstVert;
			int nNumV   = m_pFaces[i].NumVerts;

			int nTexIndex = m_pFaces[i].nTexture;
			pDevice->SetTexture( 0, 
								 m_pT[nTexIndex] );

			int nNumTris = nNumF / 3;

			m_nNumTrisDraw += nNumTris;

			pDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST,		// Type
					                        nFirstV,				// BaseVertexIndex
											0,						// MinIndex
											nNumV,					// NumVertices
											nFirstF,				// StartIndex
											nNumTris );				// PrimitiveCount
		}

	}

	void RenderLeaf(IDirect3DDevice9 * pDevice,
					 int nLeafIndex)
	{
		if( nLeafIndex > m_nNumLeafs ) assert(0 && "Passed Index");

		stLeaf * pLeaf = &m_pLeafs[nLeafIndex];

		int nFirst = pLeaf->nFirstFace;
		int nNum   = pLeaf->NumFaces;

		for(int i=0; i<nNum; i++)
		{
			int nFirstFace = m_pLeafFaces[nFirst + i];

			RenderFaces(pDevice,
						nFirstFace,
						1);
		}
	}


	void DrawPlane(IDirect3DDevice * pDevice,
				   stNode * pNode,
		           stPlane * pPlane)
	{
		// Swizzle or data into its correct format
		D3DXVECTOR3 vN;
		vN.x = pPlane->Normal[0];
		vN.y = pPlane->Normal[1];
		vN.z = pPlane->Normal[2];

		float d = pPlane->fDist;

		// Going to use the min/max to determine how long our
		// plane is
		D3DXVECTOR3 pMin = D3DXVECTOR3( (float)pNode->BBoxIMin[0],
										(float)pNode->BBoxIMin[1],
										(float)pNode->BBoxIMin[2]);

		D3DXVECTOR3 pMax = D3DXVECTOR3( (float)pNode->BBoxIMax[0],
										(float)pNode->BBoxIMax[1],
										(float)pNode->BBoxIMax[2]);

		// pA is our point the plane
		D3DXVECTOR3 pA = d * vN;
		// We'll calculate a normal perpendicular to our plane
		// by rotating vN by 90 degrees.  As were assuming our
		// planes are facing upwards in the y direction
		D3DXVECTOR3 vNN = vN;

		D3DXMATRIX matRot;
		D3DXMatrixRotationYawPitchRoll(&matRot, 0, 0, D3DX_PI/2.0f);
		D3DXVec3TransformCoord(&vNN, &vN, &matRot);

		// Draw the normal for the plane (small blue line)
		pA.z = 100;
		DrawLine3D(pDevice, pA, pA+(vN*50.0f), 0xff0f0fff);

		// Shift the plane in the diretion of the normal a bit,
		// so the plane isn't inside the wall
		pA = pA + (vN*5.0f);

		D3DXVECTOR3 vLen = pMax - pMin;
		float len = 0.7071f * D3DXVec3Length(&vLen)/2.0f;

		D3DXVECTOR3 pB = pA - (vNN*len);
		D3DXVECTOR3 pC = pA + (vNN*len);

		pC.z = 100;
		pB.z = 100;
		DrawLine3D(pDevice, pC, pB, 0xffffffff);

	}



	void CheckNode( IDirect3DDevice * pD,
				   int nNode,
				   D3DXVECTOR3 pStart,
				   D3DXVECTOR3 pEnd)
	{

		if (nNode < 0)
		{
			int nLeafIndex = -(nNode+1);

			RenderLeaf(pD, nLeafIndex);

			return;
		}


		stNode  *pNode  = &m_pNodes[nNode];
		stPlane *pPlane = &m_pPlanes[pNode->nPlane];

		DrawPlane(	pD,
					pNode,
					pPlane);

		//Vector3 vNormal = SwizzleYZ(pPlane->Normal);
		Vector3 vNormal = pPlane->Normal;
		float fStart	= vNormal.Dot(pStart)- pPlane->fDist;
		float fEnd		= vNormal.Dot(pEnd)  - pPlane->fDist;

		if (fStart >= 0 && fEnd >= 0)
		{
			CheckNode(  pD,
						pNode->nChildren[0], // front
				        pStart, pEnd);

		}
		else if (fStart < 0 && fEnd < 0)
		{
			CheckNode(  pD,
						pNode->nChildren[1], // back
				        pStart, pEnd);
		}
		else
		{
			CheckNode(  pD,
						pNode->nChildren[0], 
				        pStart, pEnd);
			CheckNode(  pD,
			    		pNode->nChildren[1], 
				        pStart, pEnd);
		}

	}



	int FindLeaf( IDirect3DDevice * pDevice,
				  const D3DXVECTOR3& Pos)
	{
		int index = 0;

		while (index >= 0) 
		{
			stNode*  pNode  = &m_pNodes[index];
			stPlane* pPlane = &m_pPlanes[pNode->nPlane];

			DrawPlane(	pDevice,
						pNode,
						pPlane);

			// Distance from point to a plane
			//Vector3 vNormal			= SwizzleYZ(pPlane->Normal);
			Vector3 vNormal			= pPlane->Normal;
			const float distance	= vNormal.Dot(Pos)- pPlane->fDist;

			//char buf[200];
			//sprintf(buf,"index:%d  Normal: %.2f,%.2f,%.2f, d:%.2f", 
			//			 index, 
			//			 vNormal.x,vNormal.y,vNormal.z, 
			//			 distance );
			//abc(buf);

			if (distance >= 0) 
			{
				index = pNode->nChildren[0];// front
			} else 
			{
				index = pNode->nChildren[1]; //back;
			}

			//sprintf(buf," -> index(%d)\n", index);
			//abc(buf);
		}

		return -index - 1;
	}


	void CheckNodeForPoint( IDirect3DDevice * pDevice,
							D3DXVECTOR3 pPoint)
	{

		int i = FindLeaf(pDevice,
						 pPoint);

		RenderLeaf(pDevice, i);

	}

	void CheckNodeForLine( IDirect3DDevice * pDevice,
							D3DXVECTOR3 pStart,
							D3DXVECTOR3 pEnd)
	{
		DrawLine3D(pDevice,
				pStart,
				pEnd,
				0xff00ffff);


		CheckNode(	pDevice, 
					0, 
					pStart,
					pEnd);
	}




							 

	void CheckMultipleLines(IDirect3DDevice * pDevice,
							 D3DXVECTOR3 pOrigin,
							 D3DXVECTOR3 vDir,
							 float fLen)
	{
	

		D3DXMATRIX mRotR, mRotL;
		D3DXMatrixRotationYawPitchRoll(&mRotR, 0, 0, D3DX_PI/8);
		D3DXMatrixRotationYawPitchRoll(&mRotL, 0, 0, -D3DX_PI/8);

		D3DXVECTOR3 vDirMiddle, vDirRight, vDirLeft;
		D3DXVec3Normalize(&vDirMiddle,	&vDir);
		D3DXVec3Normalize(&vDirRight,	&vDir);
		D3DXVec3Normalize(&vDirLeft,	&vDir);

		D3DXVec3TransformCoord(&vDirRight, &vDirRight, &mRotR);
		D3DXVec3TransformCoord(&vDirLeft, &vDirLeft, &mRotL);

		vDirMiddle	= pOrigin	+ (vDirMiddle*fLen);
		vDirRight	= pOrigin	+ (vDirRight*fLen);
		vDirLeft	= pOrigin	+ (vDirLeft*fLen);

		CheckNode(	pDevice, 
					0, 
					pOrigin,
					vDirMiddle);

		CheckNode(	pDevice, 
					0, 
					pOrigin,
					vDirLeft);

		CheckNode(	pDevice, 
					0, 
					pOrigin,
					vDirRight);

		DrawLine3D(pDevice,
				pOrigin,
				vDirMiddle,
				0xff00ffff);

		DrawLine3D(pDevice,
				pOrigin,
				vDirLeft,
				0xff00ffff);

		DrawLine3D(pDevice,
				pOrigin,
				vDirRight,
				0xff00ffff);

	}


	//See if one cluster is visible from another
	bool isClusterVisible(int visCluster, int testCluster)
	{
		if( visCluster		< 0) return 0;
		if( testCluster		< 0) return 0;
		if( m_VisData.pBits==NULL) return 0;


		int bytesPerCluster = m_VisData.nSizeVectorInBytes;

		byte visSet = m_VisData.pBits[(visCluster * bytesPerCluster) + (testCluster >> 3)];

		return (visSet & (1 << (testCluster & 7))) != 0;

		// MACRO VERSION
		// #define M_ISVISIBLE(VD,x,y) ( ( (1<<((y)&7)) & (VD).vecs[(x)*(VD).sz_vecs+((y)>>3)] ) != 0   )

	}


	void DrawFace(IDirect3DDevice9 * pDevice,
				  int nFace)
	{

		pDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
		pDevice->SetTextureStageState(0,D3DTSS_COLOROP,	D3DTOP_SELECTARG2);
		pDevice->SetTextureStageState(0,D3DTSS_COLORARG1,	D3DTA_DIFFUSE);
		pDevice->SetTextureStageState(0,D3DTSS_COLORARG2,	D3DTA_TEXTURE);
		
		pDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
		pDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE); 

		pDevice->SetIndices(m_pIB);

		pDevice->SetStreamSource(0, m_pVB, 0, sizeof(stBSPVertex));

		pDevice->SetFVF( m_FVF_Format );

		// If its not a tri mesh
		if( m_pFaces[nFace].nFaceType != 1)return;
		
		int nFirstF = m_pFaces[nFace].nFirstMeshVerts;
		int nNumF   = m_pFaces[nFace].NumMeshVerts;

		int nFirstV = m_pFaces[nFace].FirstVert;
		int nNumV   = m_pFaces[nFace].NumVerts;

		int nTexIndex = m_pFaces[nFace].nTexture;

		pDevice->SetTexture( 0, m_pT[nTexIndex] );

		int nNumTris = nNumF / 3;

		/*
		pDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST,		// Type
					                    nFirstV,				// BaseVertexIndex
										0,						// MinIndex
										nNumV,					// NumVertices
										nFirstF,				// StartIndex
										nNumTris );				// PrimitiveCount
		*/

		m_nNumTrisDraw += (m_pFaces[nFace].NumMeshVerts/3);

		pDevice->DrawIndexedPrimitive(	D3DPT_TRIANGLELIST , 
										m_pFaces[nFace].FirstVert, 
										0, 
										m_pFaces[nFace].NumVerts,
										m_pFaces[nFace].nFirstMeshVerts, 
										(m_pFaces[nFace].NumMeshVerts/3) );

	}

	//void abc(char *str)
	//{
	//	//FILE *fp = fopen("output.txt", "a+");
	//	//fprintf(fp, "%s\n", str);
	//	//fclose(fp);
	//}
protected:
	void TurnFaceOn(int nFace)
	{
		m_bFaceVisibility[nFace] = true;
	}

	void TurnAllFacesOff()
	{
		for(int i=0; i<m_nNumFaces; i++)
			m_bFaceVisibility[i]=false;
	}

	void RenderAllFaces(IDirect3DDevice * pDevice)
	{
		for(int i=0; i<m_nNumFaces; i++)
		{
			if( !m_bFaceVisibility[i] ) continue;

			DrawFace( pDevice, i );
		}
	}

public:

	void CheckClusterPoint(IDirect3DDevice * pDevice,
						   D3DXVECTOR3 pPoint )
	{
		//If cluster is negative, the leaf is outside the map or otherwise invalid.
		int indx = FindLeaf(pDevice,
						 pPoint);

		RenderLeaf(pDevice, indx);

		int nCluster = m_pLeafs[indx].nCluster;

		//Check if where outside our world!
		if( nCluster == -1 )return; 

		TurnAllFacesOff();

		for(int i=0; i<m_nNumLeafs; i++)
		{
			int nOtherCluster = m_pLeafs[i].nCluster;

			if( isClusterVisible( nCluster, nOtherCluster ) )
			{
				//char buf[200];
				//sprintf(buf, "Index:%d  (Cluster:%d, OtherClus:%d)  Leaf: %d  [Leaffaces(%d %d)]", 
				//		        i, nCluster, nOtherCluster,
				//				i, m_pLeafs[i].nFirstFace, m_pLeafs[i].NumFaces);
				//abc(buf);

				for(int f = 0; f < m_pLeafs[i].NumFaces; f++)
				{
//					DrawFace( pDevice,
//						m_pLeafFaces[m_pLeafs[i].nFirstFace+f] );

					TurnFaceOn( m_pLeafFaces[m_pLeafs[i].nFirstFace+f] );

					//sprintf(buf, "%.2d  LeafFace: %.2d", (i+f), m_pLeafFaces[m_pLeafs[i].nFirstFace+f]);
					//abc(buf);

				}
			}
		}

		RenderAllFaces(pDevice);
	}
};//End CBSP class






#endif // BSP_H